/* Loads a file and sends it to each client that connects.
*/

#include "Dasyne.hpp"

using namespace std;
using namespace libdasyne;
class Client;

//False if we should stop the server
static bool allOK = true;

//The file to send to clients
static DBuffer theFile;

//The size of theFile in bytes
static DBuffer fileSize;

//A list of clients who've connected, in a proper server we'd delete these when they
//disconnect
static vector<Client*> theClients;

//The port number to communicate over
static const uint16_t PORT_NUMBER = 7890;

bool loadFile(string fileName);

class Client : public DClient
{
    public:
        //constructor, passes the network engine and an existing connection to DClient
        Client(DEngine *newNetEngine,
               boost::asio::ip::tcp::socket *newSocket) : DClient(newNetEngine, newSocket)
        {
        }//constructor

    private:
        //Event handler called when we receive data from this client
        void receiveHandler(DBuffer &theData)
        {
            cout << "Received: " << theData.toString() << endl;
        }//receiveHandler

        //Event handler called when a network error occurs and the client
        //disconnects.
        void errorHandler(std::string message)
        {
            cout << "Client disconnected: " << message << endl;
        }//errorHandler
};//Client

class FileServer : public DServer
{
    public:
        //Constructor, takes a pointer to a DEngine and passes it to the
        //parent DServer object
        FileServer(DEngine *newNetEngine) : DServer(newNetEngine)
        {
        }//constructor

    private:
        //Event handler, called when a client connects
        void connectHandler(boost::asio::ip::tcp::socket *newSocket)
        {
            Client *newClient = new Client(netEngine, newSocket);

            cout << "A client connected from " << newClient->getRemoteAddress() << endl;
            theClients.push_back(newClient);
            newClient->send(fileSize);
            newClient->send(theFile);
        }//connectHandler

        //Event handler, called when there's an error accepting connections
        void errorHandler(string errMsg)
        {
            cout << errMsg << endl;
            allOK = false;
        }//errorHandler
};//FileServer

int main(int argc,
         char *argv[])
{
    DEngine netEngine;//calls event handlers for networking classes
    FileServer *myServer = new FileServer(&netEngine);//listens for client connections
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Verify that a file name was given ###
    if (argc != 2)
    {
        cout << "Usage: fileserver <file_name>" << endl;
        allOK = false;
    }//if argc

    //### Bind the listen socket ###
    if (allOK)
    {
        if (!myServer->bind(libdasyne::DA_IPV4, PORT_NUMBER))
        {
            allOK = false;
            cout << "Couldn't bind: " << myServer->getErrMsg() << endl;
        }//if !bind()
    }//if allOK

    //### Start listening for connections ###
    if (allOK)
    {
        if (!myServer->startListening())
        {
            allOK = false;
            cout << "Couldn't listen: " << myServer->getErrMsg() << endl;
        }//if !startListening()
    }//if allOK

    //### Load the file ###
    if (allOK)
    {
        if (loadFile(argv[1]))
            cout << "Serving " << argv[1] << ", hit Ctrl-C to halt" << endl;
        else
            allOK = false;
    }//if allOK

    //### Periodically call event handlers ###
    while (allOK)
    {
        eventsRun = netEngine.update(1000);
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg() << endl;
            allOK = false;
        }//if eventsRun

        //Sleep if we have nothing to do
        else if (eventsRun == 0)
            DEngine::sleep(1);
    }//while allOK

    //### Delete the objects that use DEngine before it's deleted ###
    for (unsigned int i = 0; i < theClients.size(); i++)
        delete theClients.at(i);
    theClients.clear();

    //### Free memory ###
    delete myServer;
}//main

bool loadFile(string fileName)
{
    ifstream infile;//obtains binary input from the file
    uint64_t numBytes = 0;//the size of the file in bytes
    uint8_t oneByte = 0;//a single byte from the file
    bool result = true;//false if we couldn't load the file.  Either file not found,
                      //or we ran out of memory

    //### Open the file ###
    infile.open(fileName.c_str(), ios::binary);
    if (!infile.is_open())
    {
        result = false;
        cout << "Couldn't open " << fileName << endl;
    }//if !is_open()

    //### Load the file ###
    if (result)
    {
        cout << "Loading " << fileName << "... ";

        oneByte = infile.get();
        while (!infile.eof())
        {
            theFile.appendByte(oneByte);
            oneByte = infile.get();
        }//while !eof()

        cout << "complete." << endl;
        cout << "File size is " << theFile.size() << " bytes" << endl;
        cout << endl;
    }//if result

    //### Store the length in fileSize ###
    if (result)
    {
        numBytes = theFile.size();
        fileSize.appendVariable(numBytes);
    }//if result

    //### Close the file ###
    if (infile.is_open())
        infile.close();

    return result;
}//loadFile
