/* Shows how to connect and receive data with the DDispatcher class. We
   receive three types of messages from the server, and the DDispatcher automatically
   calls the proper event handler, based on the message type.
*/

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;

//The port number to communicate over
const uint16_t PORT_NUMBER = 5137;

//The number of messages we've received from the server
static unsigned long int received1Messages = 0;
static unsigned long int received219Messages = 0;
static unsigned long int received220Messages = 0;

//False if we encountered an error and should stop the program
static bool allOK = true;

class MessageClient : public DDispatcher
{
    public:
        //Constructor, we pass the DEngine to the parent DMessageClient class
        MessageClient(DEngine *newNetEngine) : DDispatcher(newNetEngine)
        {
            addHandler(1, boost::bind(&MessageClient::oneHandler, this, _1));
            addHandler(2, 19, boost::bind(&MessageClient::twoNineteenHandler, this, _1));
            addHandler(2, 20, boost::bind(&MessageClient::twoTwentyHandler, this, _1));
        }//constructor

        //Connect to the host on port 'PORT_NUMBER'
        void connectToHost(string hostname)
        {
            startConnect(hostname, PORT_NUMBER);
        }//connectToHost

        //Event handler, called when the connect operation finishes
        void connectHandler(string message,
                            libdasyne::NetworkCode status)
        {
            if (status == libdasyne::DA_SUCCESS)
                cout << "Connected to " << getRemoteAddress() << endl;
            else
            {
                cout << message << endl;
                allOK = false;
            }//else status
        }//connectHandler

        //Event handler, called when we receive a message of type 1
        void oneHandler(DBuffer &theMessage)
        {
            unsigned int byteIndex = 0;//the index of a byte in theMessage

            //### Verify that we got a complete message ###
            if (theMessage.size() != 256)
            {
                cout << "Expecting a type 1 message 256 bytes long, got "
                     << theMessage.size() << " bytes" << endl;
                allOK = false;
            }//if size()

            //### Check each byte in the message ###
            while ((allOK) && (byteIndex < theMessage.size()))
            {
                if (theMessage.getByte(byteIndex) != byteIndex)
                {
                    cout << "Message 1 expected " << byteIndex
                         << ", got " << (int)theMessage.getByte(byteIndex) << endl;
                                             
                    allOK = false;
                }//if theMessage

                byteIndex++;
            }//while allOK && byteIndex

            //### Record that we've received a message ###
            if (allOK)
                received1Messages++;
        }//messageHandler

        //Event handler, called when we receive a message of type 2, subtype 19
        void twoNineteenHandler(DBuffer &theMessage)
        {
            unsigned int byteIndex = 0;//the index of a byte in theMessage
            
            //### Verify that we got a complete message ###
            if (theMessage.size() != 50)
            {
                cout << "Expecting a type 2-19 message 50 bytes long, got "
                     << theMessage.size() << " bytes" << endl;
                allOK = false;            
            }//if size
            
            //### Check each byte in the message ###
            while ((allOK) && (byteIndex < theMessage.size()))
            {
                if (theMessage.getByte(byteIndex) != byteIndex * 2)
                {
                    cout << "Message 219 expected " << byteIndex * 2
                         << ", got " << (int)theMessage.getByte(byteIndex) << endl;
                                             
                    allOK = false;
                }//if theMessage

                byteIndex++;
            }//while allOK && byteIndex
            
            //### Record that we've received a message ###
            if (allOK)
                received219Messages++;                  
        }//twoNineteenHandler

        //Event handler, called when we receive a message of type 2, subtype 20
        void twoTwentyHandler(DBuffer &theMessage)
        {
            unsigned int byteIndex = 0;//the index of a byte in theMessage
            
            //### Verify that we got a complete message ###
            if (theMessage.size() != 25)
            {
                cout << "Expecting a type 2-20 message 25 bytes long, got "
                     << theMessage.size() << " bytes" << endl;
                allOK = false;            
            }//if size
            
            //### Check each byte in the message ###
            while ((allOK) && (byteIndex < theMessage.size()))
            {
                if (theMessage.getByte(byteIndex) != byteIndex * 3)
                {
                    cout << "Message 220 expected " << byteIndex * 3
                         << ", got " << (int)theMessage.getByte(byteIndex) << endl;
                                             
                    allOK = false;
                }//if theMessage

                byteIndex++;
            }//while allOK && byteIndex
            
            //### Record that we've received a message ###
            if (allOK)
                received220Messages++;                   
        }//twoTwentyHandler

        //Event handler, called when a network error occurs
        void errorHandler(string message)
        {
            cout << "Client disconnected: " << message << endl;
            allOK = false;
        }//errorHandler

        //Event handler called when we get a corrupted message from the server
        void corruptHandler(string message)
        {
            cout << message << endl;
        }//corruptHandler
};//MessageClient

void showMessages(void)
{
    cout << "Received " << received1Messages << " type 1 messages" << endl;
    cout << "         " << received219Messages << " type 219 messages" << endl;
    cout << "         " << received220Messages << " type 220 messages" << endl;
    cout << endl;
}//showMessages

void stopClient(void)
{
   cout << "Final results:" << endl;
   showMessages();
   allOK = false;
}//stopClient

int main(int argc,
         char *argv[])
{
    DEngine *netEngine = new DEngine;//handles network events
    MessageClient *client = new MessageClient(netEngine);//sends and receives text messages
    DTimer msgTimer(netEngine);//periodically show how many messages we've received
    DTimer clientTimer(netEngine);//stop the program after a specified interval
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Periodically show how many messages were received ###
    msgTimer.setInterval(3000);
    msgTimer.setHandler(&showMessages);
    msgTimer.start();

    //### Stop the client after 18 seconds ###
    clientTimer.setInterval(18000);
    clientTimer.setHandler(&stopClient);
    clientTimer.start();

    //### Verify that the proper number of arguments were given ###
    if (argc != 2)
    {
        cout << "Usage: msgclient <host name>" << endl;
        cout << endl;
        allOK = false;
    }//if argc

    //### Start connecting to the server ###
    else
        client->connectToHost(argv[1]);

    while (allOK)
    {
        //### Dispatch any queued events ###
        eventsRun = netEngine->update(1000);
        if (eventsRun < 0)
        {
            cout << netEngine->getErrMsg() << endl;
            allOK = false;
        }//if eventsRun

        //### Sleep if we have nothing to do ###
        else if (eventsRun == 0)
            DEngine::sleep(1);
    }//while allOK

    //### Delete the classes that use netEngine before deleting netEngine ###
    delete client;
    delete netEngine;

    return 0;
}//main
