/* Tests the throughput of the link to a tserver program. See the
  ThroughputServer.cpp file for details.
*/

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;

//The port number to communicate over
const uint16_t PORT_NUMBER = 5127;

//The total number of kilobytes received
static double kbReceived = 0;

//The number of kilobytes to allocate in our receive buffer
static int receiveBufferSize = 0;

//False if we encountered an error and should stop the program
static bool allOK = true;

//Shows tclient's command-line usage
void showUsage(void);

class TpClient : public DClient
{
    public:
        //Constructor, we pass the DEngine to the parent DClient class
        TpClient(DEngine *newNetEngine,
                 unsigned int myBufferSize) : DClient(newNetEngine, myBufferSize)
        {
        }//constructor

        //Connect to the host on port 'PORT_NUMBER'
        void connectToHost(string hostname)
        {
            startConnect(hostname, PORT_NUMBER);
        }//connectToHost

        //Event handler, called when the connect operation finishes
        void connectHandler(string message,
                            libdasyne::NetworkCode status)
        {
            if (status == libdasyne::DA_SUCCESS)
            {
                cout << "Connected to " << getRemoteAddress() << endl;
                cout << "My address: " << getLocalAddress() << endl;
            }//if status
            else
            {
                cout << message << endl;
                allOK = false;
            }//else status
        }//connectHandler

        //Event handler, called when we receive data
        void receiveHandler(DBuffer &receivedBytes)
        {
            kbReceived += receivedBytes.size() / 1024.0;
        }//receiveHandler

        //Event handler, called when a network error occurs
        void errorHandler(string message)
        {
            cout << "Connection lost." << endl;
            allOK = false;
        }//errorHandler
};//TpClient

int main(int argc,
         char *argv[])
{
    //mtrace();
    DEngine *netEngine = new DEngine;//handles network events
    TpClient *client = NULL;//sends and receives text messages
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Verify that the proper number of arguments were given ###
    if (argc != 3)
    {
        showUsage();
        allOK = false;
    }//if argc

    //### Extract the buffer size parameter ###
    if (allOK)
    {
        if (!DEngine::strToInt(argv[2], receiveBufferSize))
        {
            showUsage();
            allOK = false;
        }//if !strToInt()
    }//if allOK

    //### Create the client and connect ###
    if (allOK)
    {
        if (receiveBufferSize > 0)
        {
            client = new TpClient(netEngine, receiveBufferSize * 1024);
            client->connectToHost(argv[1]);
        }//if receiveBufferSize
        else
        {
            showUsage();
            allOK = false;
        }//else receiveBufferSize
    }//if allOK

    while (allOK)
    {
        //### Dispatch any queued events ###
        eventsRun = netEngine->update(-1);
        if (eventsRun < 0)
        {
            cout << netEngine->getErrMsg() << endl;
            allOK = false;
        }//if eventsRun

        //### Sleep if we have nothing to do ###
        else if (eventsRun == 0)
            DEngine::sleep(1);
    }//while allOK

    //### Output results ###
    if (kbReceived > 1048576)
        cout << "Received " << kbReceived / 1048576.0 << " GB" << endl;
    else if (kbReceived > 1048576)
        cout << "Received " << kbReceived / 1024.0 << " MB" << endl;
    else
        cout << "Received " << kbReceived << " KB" << endl;

    //### Delete the classes that use netEngine before deleting netEngine ###
    delete client;
    delete netEngine;

    return 0;
}//main

void showUsage(void)
{
    cout << "Usage: tclient <host_name> <buffer_KB>" << endl;
    cout << endl;
}//showUsage
