/*
 * Copyright (C) 2005-2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/*
  The following copyright notice applies to the crcinit, reflect, and crc32 functions.
  These were derived from Michael Barr's original algorithm.

 *#####################################################################
 *
 * Filename:    crc.c
 *
 * Description: Slow and fast implementations of the CRC standards.
 *
 * Notes:       The parameters for each supported CRC standard are
 *                              defined in the header file crc.h.  The implementations
 *                              here should stand up to further additions to that list.
 *
 *
 * Copyright (c) 2000 by Michael Barr.  This software is placed into
 * the public domain and may be used for any purpose.  However, this
 * notice must not be changed or removed and no warranty is either
 * expressed or implied by its publication or distribution.
 #####################################################################*/

#include "DBuffer.hpp"
#include "DBufferIter.hpp"

using namespace std;
using namespace libdasyne;

uint32_t *DBuffer::crcTable = crcInit();

DBuffer::DBuffer(void)
{
    errMsg = "AllOK";
}//constructor

DBuffer::DBuffer(const DBuffer &source) : DBytes(source)
{
    errMsg = source.errMsg;
}//copy constructor

DBuffer::~DBuffer(void)
{
}//destructor

DBuffer& DBuffer::operator= (const DBuffer &source)
{
    if (this != &source)
    {
        clear();
        append(source);
        errMsg = source.errMsg;
    }//if this

    return *this;
}//=

uint32_t* DBuffer::crcInit(void)
{
    uint32_t *result = NULL;//the CRC table to return
    uint32_t crcWidth = 8 * sizeof(uint32_t);
    uint32_t crcTopBit = (1 << (crcWidth - 1));
    uint32_t remainder = 0;
    int32_t dividend = 0;
    uint8_t bit = 0;

    result = new uint32_t[256];

    for (dividend = 0; dividend < 256; ++dividend)
    {
        remainder = dividend << (crcWidth - 8);

        for (bit = 8; bit > 0; --bit)
        {
            if (remainder & crcTopBit)
                remainder = (remainder << 1) ^ 0x04C11DB7;
            else
                remainder = (remainder << 1);
        }//for bit

        result[dividend] = remainder;
    }//for dividend

    return result;
}//crcInit

uint32_t DBuffer::reflect(uint32_t data,
                          uint8_t nBits) const
{
    uint32_t reflection = 0x00000000;
    uint8_t bit;

    for (bit = 0; bit < nBits; ++bit)
    {
        if (data & 0x01)
            reflection |= (1 << ((nBits - 1) - bit));
        data = (data >> 1);
    }//for bit

    return (reflection);
}//reflect

uint32_t DBuffer::crc32(void) const
{
    DBufferIter dIter;//iterates through the bytes in this DBuffer
    uint32_t crcWidth = 8 * sizeof(uint32_t);
    uint32_t remainder = 0xFFFFFFFF;
    uint8_t data = 0;

    dIter.setBuffer(this);
    dIter.setIndex(0);

    while (!dIter.finished())
    {
        data = ((uint8_t) reflect((dIter.value()), 8)) ^ (remainder >> (crcWidth - 8));
        remainder = crcTable[data] ^ (remainder << 8);

        dIter.next();
    }//byte

    return (((uint32_t) reflect((remainder), crcWidth)) ^ 0xFFFFFFFF);
}//crc32

DBuffer DBuffer::compress(int compressionLevel)
{
    DBuffer result;//the compressed version of the data
    z_stream defStream;//compresses the data using zlib
    uint8_t outputBuffer[ZLIB_BUFFER_SIZE];//output from zlib
    bool allOK = true;//false if an error occurred

    if (size() > 0)
    {
        //### Set up the z_stream ###
        if (!initializeDeflation(defStream, compressionLevel))
            allOK = false;

        //### Feed the uncompressed data into the stream ###
        if (allOK)
        {
            if (!runDeflation(defStream, result, outputBuffer))
                allOK = false;
        }//if allOK

        //### Extract the compressed data from zlib ###
        if (allOK)
        {
            if (!finishDeflation(defStream, result, outputBuffer))
                allOK = false;
        }//if allOK

        //### Close the stream ###
        if (allOK)
        {
            if (!endDeflation(defStream))
                allOK = false;
        }//if allOK

        //### Obtain any error messages from zlib ###
        if (!allOK)
        {
            if (defStream.msg != NULL)
                errMsg.append(defStream.msg);
        }//if !allOK
    }//if size()
    else
    {
        allOK = false;
        errMsg = "No data to compress.";
    }//else if size()

    //### Return an empty DBuffer on errors ###
    if (!allOK)
        result.clear();

    return result;
}//compress

bool DBuffer::initializeDeflation(z_stream& defStream,
                                  int compressionLevel)
{
    bool result = true;//false if errors occur
    int initResult = 0;//the return code from deflateInit()

    defStream.zalloc = (alloc_func)0;
    defStream.zfree = (free_func)0;
    defStream.opaque = (voidpf)0;

    initResult = deflateInit(&defStream, compressionLevel);
    if (initResult != Z_OK)
    {
        if (initResult == Z_MEM_ERROR)
            errMsg = "deflateInit(): Insufficient memory";
        else if (initResult == Z_STREAM_ERROR)
            errMsg = "deflateInit(): invalid compression level";
        else if (initResult == Z_VERSION_ERROR)
            errMsg = "deflateInit(): incompatible library version";

        result = false;
    }//if initResult

    return result;
}//initializeDeflation

bool DBuffer::runDeflation(z_stream& defStream,
                           DBuffer &compressedBytes,
                           uint8_t outputBuffer[])
{
    DBufferIter dIter;//iterates through the stored bytes
    uint8_t inputBuffer[ZLIB_BUFFER_SIZE];//input to be compressed
    int deflateResult = 0;//result of calling deflate()
    bool result = true;//false if deflate() gives us an error

    dIter.setBuffer(this);
    dIter.setIndex(0);

    defStream.avail_in = fillCompressionInput(inputBuffer, dIter);
    defStream.next_in = (Bytef*)inputBuffer;

    defStream.avail_out = ZLIB_BUFFER_SIZE;
    defStream.next_out = (Bytef*)outputBuffer;

    while (defStream.avail_in > 0)
    {
        deflateResult = deflate(&defStream, Z_NO_FLUSH);

        if (deflateResult != Z_OK)
        {
            if (deflateResult == Z_STREAM_ERROR)
            {
                errMsg = "deflate(): Stream was inconsistent";
                result = false;
            }//if deflateResult

            else if (deflateResult == Z_BUF_ERROR)
            {
                if (defStream.avail_out == 0)
                {
                    emptyCompressionOutput(outputBuffer, ZLIB_BUFFER_SIZE, compressedBytes);
                    defStream.avail_out = ZLIB_BUFFER_SIZE;
                    defStream.next_out = (Bytef*)outputBuffer;
                }//if avail_out
                else
                {
                    errMsg = "deflate(): No progress was possible";
                    result = false;
                }//else if avail_out
            }//else if deflateResult
        }//if deflateResult

        if (defStream.avail_in == 0)
        {
            defStream.avail_in = fillCompressionInput(inputBuffer, dIter);
            defStream.next_in = (Bytef*)inputBuffer;
        }//if avail_in
    }//while avail_in

    return result;
}//runDeflation()

bool DBuffer::finishDeflation(z_stream& defStream,
                              DBuffer &compressedBytes,
                              uint8_t outputBuffer[])
{
    bool result = true;//false if deflate gave us errors
    bool finished = false;//true if we've 'finished' the deflation
    int deflateResult = 0;//return code from deflate()

    defStream.avail_in = 0;

    while ((!finished) && (result))
    {
        if (defStream.avail_out == 0)
        {
            emptyCompressionOutput(outputBuffer, ZLIB_BUFFER_SIZE, compressedBytes);
            defStream.avail_out = ZLIB_BUFFER_SIZE;
            defStream.next_out = (Bytef*)outputBuffer;
        }//if avail_out

        deflateResult = deflate(&defStream, Z_FINISH);

        if (deflateResult == Z_STREAM_END)
            finished = true;
        else
        {
            if (deflateResult != Z_OK)
            {
                if (deflateResult == Z_STREAM_ERROR)
                    errMsg = "deflate(): Stream was inconsistent";
                else if (deflateResult == Z_BUF_ERROR)
                    errMsg = "deflate(): No progress was possible";

                result = false;
            }//if deflateResult
        }//else if deflateResult
    }//while !finished && result

    if (result)
        emptyCompressionOutput(outputBuffer, defStream.total_out - compressedBytes.size(),
                               compressedBytes);

    return result;
}//finishDeflation

bool DBuffer::endDeflation(z_stream& defStream)
{
    bool result = true;//false if deflateEnd() gave us errors
    int deflateResult = deflateEnd(&defStream);

    if (deflateResult != Z_OK)
    {
        if (deflateResult == Z_STREAM_ERROR)
            errMsg = "deflateEnd(): The stream state was inconsistent.";
        if (deflateResult == Z_DATA_ERROR)
            errMsg = "deflateEnd(): The stream was freed prematurely.  Some data may be lost.";

        result = false;
    }//if deflateResult

    return result;
}//endDeflation

int DBuffer::fillCompressionInput(uint8_t inputBuffer[],
                                  DBufferIter &dIter)
{
    int bufPos = 0;//our current position in inputBuffer[]

    while ((!dIter.finished()) && (bufPos < ZLIB_BUFFER_SIZE))
    {
        inputBuffer[bufPos] = dIter.value();

        dIter.next();
        bufPos++;
    }//while charPos

    return bufPos;
}//fillCompressionInput

void DBuffer::emptyCompressionOutput(uint8_t outputBuffer[],
                                     int bufferLength,
                                     DBuffer &compressedBytes)
{
    compressedBytes.appendArray(outputBuffer, bufferLength);
}//emptyCompressionOutput

DBuffer DBuffer::decompress(void)
{
    z_stream infStream;//the decompression stream
    DBuffer result;//the decompressed data to return
    int inflateResult = 0;//the result from calling inflateEnd()
    bool allOK = true;//false on errors

    //### Verify that we have some data ###
    if (size() <= 0)
    {
        allOK = false;
        errMsg = "No data to decompress.";
    }//if size()

    //### Initialize the inflation stream ###
    if (allOK)
    {
        if (!initializeInflation(infStream))
            allOK = false;
    }//if allOK

    //### Decompress the data ###
    if (allOK)
    {
        if (!runInflation(infStream, result))
            allOK = false;
    }//if allOK

    //### Close the stream ###
    if (allOK)
    {
        inflateResult = inflateEnd(&infStream);

        if (inflateResult == Z_STREAM_ERROR)
        {
            errMsg = "inflateEnd(): Inconsistent inflate stream";
            allOK = false;
        }//if inflateResult
    }//if allOK

    //### Check for any zlib error messages ###
    if (!allOK)
    {
        if (infStream.msg != NULL)
            errMsg.append(infStream.msg);

        result.clear();
    }//if !allOK

    return result;
}//decompress

bool DBuffer::initializeInflation(z_stream& infStream)
{
    bool result = true;//false if inflateInit() gave us errors
    int inflateResult = 0;//return code from inflate()

    infStream.zalloc = (alloc_func)0;
    infStream.zfree = (free_func)0;
    infStream.opaque = (voidpf)0;

    inflateResult = inflateInit(&infStream);

    if (inflateResult != Z_OK)
    {
        if (inflateResult == Z_MEM_ERROR)
            errMsg = "inflateInit(): Insufficient memory.";
        else if (inflateResult == Z_VERSION_ERROR)
            errMsg = "inflateInit(): incompatible zlib version.";

        result = false;
    }//else if inflateResult

    return result;
}//initializeInflation

bool DBuffer::runInflation(z_stream& infStream,
                           DBuffer &extractedBytes)
{
    DBufferIter dIter;//iterates through the bytes being decompressed
    int inflateResult = 0;//return code from inflate()
    uint8_t inputBuffer[ZLIB_BUFFER_SIZE];//input from the deque
    uint8_t outputBuffer[ZLIB_BUFFER_SIZE];//output from zlib
    bool streamEnd = false;//true if we hit the end of the stream
    bool result = true;//false if inflate() gives us errors

    dIter.setBuffer(this);
    dIter.setIndex(0);

    infStream.avail_in = fillCompressionInput(inputBuffer, dIter);
    infStream.next_in = (Bytef*)inputBuffer;

    infStream.avail_out = ZLIB_BUFFER_SIZE;
    infStream.next_out = outputBuffer;

    while ((!streamEnd) && (result))
    {
        if (infStream.avail_out == 0)
        {
            emptyCompressionOutput(outputBuffer, ZLIB_BUFFER_SIZE, extractedBytes);
            infStream.avail_out = ZLIB_BUFFER_SIZE;
            infStream.next_out = (Bytef*)outputBuffer;
        }//if avail_out
        if (infStream.avail_in == 0)
        {
            infStream.avail_in = fillCompressionInput(inputBuffer, dIter);
            infStream.next_in = (Bytef*)inputBuffer;
        }//if avail_in

        inflateResult = inflate(&infStream, Z_NO_FLUSH);

        if (inflateResult == Z_STREAM_END)
            streamEnd = true;
        else
        {
            if (inflateResult != Z_OK)
            {
                if (inflateResult == Z_NEED_DICT)
                    errMsg.append("inflate(): Preset dictionary is required.");
                else if (inflateResult == Z_DATA_ERROR)
                    errMsg.append("inflate(): Input data corrupted");
                else if (inflateResult == Z_STREAM_ERROR)
                    errMsg.append("inflate(): Inflate stream structure was inconsistent.");
                else if (inflateResult == Z_MEM_ERROR)
                    errMsg.append("inflate(): Insufficient memory");
                else if (inflateResult == Z_BUF_ERROR)
                    errMsg.append("inflate(): No progress was possible");

                result = false;
            }//if inflateResult
        }//else if inflateResult
    }//while !streamEnd && result

    if (result)
        emptyCompressionOutput(outputBuffer, infStream.total_out - extractedBytes.size(),
                               extractedBytes);

    return result;
}//runInflation

string DBuffer::toString(void) const
{
    DBufferIter dataIter;//iterates through the elements in byteData
    string result;//the entire contents of this DBytes, in string form

    dataIter.setBuffer(this);
    dataIter.setIndex(0);

    while (!dataIter.finished())
    {
        result += dataIter.value();
        dataIter.next();
    }//while !finished()

    return result;
}//toString

void DBuffer::appendString(const std::string &source)
{
    const char *cString = source.c_str();

    appendArray((const uint8_t*)cString, source.size());
}//appendString

string DBuffer::getErrMsg(void) const
{
    return errMsg;
}//getErrMsg
