/*
 * Copyright (C) 2005-2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DBytes.hpp"
#include "DSafeArray.hpp"

using namespace std;
using namespace libdasyne;

#ifdef DA_DEBUG_MODE
    long int DBytes::numInstances = 0;
#endif

DBytes::DBytes(void)
{
    #ifdef DA_DEBUG_MODE
        numInstances++;
        arraysAdded = 0;
    #endif

    startFilledBytes = 0;
    endFilledBytes = 0;

    clear();
}//constructor

DBytes::DBytes(const DBytes &source)
{
    #ifdef DA_DEBUG_MODE
        numInstances++;
        arraysAdded = 0;
    #endif

    startFilledBytes = 0;
    endFilledBytes = 0;

    append(source);
}//copy constructor

DBytes::~DBytes(void)
{
    #ifdef DA_DEBUG_MODE
        numInstances--;
        arraysAdded -= byteData.size();
    #else
        for (unsigned int i = 0; i < byteData.size(); i++)
            delete [] byteData.at(i);
    #endif

    byteData.clear();

    #ifdef DA_DEBUG_MODE
        if (arraysAdded != 0)
        {
            DEngine::addFatal("DBytes::~DBytes(): leaked ");
            DEngine::addFatal(DEngine::intToStr(arraysAdded));
            DEngine::fatalError(" SafeArrays.");
        }//if arraysAdded
    #endif
}//destructor

DBytes& DBytes::operator= (const DBytes &source)
{
    if (this != &source)
    {
        clear();
        append(source);
    }//if this

    return *this;
}//=

void DBytes::setByte(long int pos,
                     uint8_t newByte)
{
    uint32_t dataIndex = 0;//the index in byteData to access
    unsigned int arrayIndex = 0;//the index in the array to access

    calcIndexes(pos, dataIndex, arrayIndex);

    byteData.at(dataIndex)[arrayIndex] = newByte;
}//setByte

uint8_t DBytes::getByte(long int pos) const
{
    uint32_t dataIndex = 0;//the index in byteData to access
    unsigned int arrayIndex = 0;//the index in the array to access

    calcIndexes(pos, dataIndex, arrayIndex);

    return byteData.at(dataIndex)[arrayIndex];
}//getByte

void DBytes::calcIndexes(long int pos,
                         uint32_t &dataIndex,
                         unsigned int &arrayIndex) const
{
    pos += (ARRAY_SIZE - startFilledBytes);

    if (pos > 0)
    {
        dataIndex = pos / ARRAY_SIZE;
        arrayIndex = pos % ARRAY_SIZE;
    }//if pos

    if (dataIndex == byteData.size() - 1)
    {
        if (byteData.size() == 1)
        {
            if (static_cast<int>(arrayIndex) < ARRAY_SIZE - startFilledBytes)
            {
                DEngine::addFatal("DBytes::calcIndexes(): Invalid array index (start): ");
                DEngine::fatalError(DEngine::intToStr(arrayIndex));
            }//if arrayIndex
        }//if size()
        else
        {
            if (arrayIndex >= endFilledBytes)
            {
                DEngine::addFatal("DBytes::calcIndexes(): Invalid array index (end): ");
                DEngine::fatalError(DEngine::intToStr(arrayIndex));
            }//if arrayIndex
        }//else size()
    }//if dataIndex
}//calcIndexes

void DBytes::clear(void)
{
    #ifndef DA_DEBUG_MODE
        for (unsigned int i = 0; i < byteData.size(); i++)
            delete byteData.at(i);
    #else
        arraysAdded -= byteData.size();
    #endif

    byteData.clear();

    startFilledBytes = 0;
    addElement();
}//clear

unsigned long int DBytes::size(void) const
{
    unsigned long int result = 0;//the size to return

    if (byteData.size() >= 2)
        result = (byteData.size() - 2) * ARRAY_SIZE;

    result += startFilledBytes + endFilledBytes;

    return result;
}//size

void DBytes::appendByte(uint8_t source)
{
    if ((endFilledBytes >= ARRAY_SIZE) || (byteData.size() == 1))
        addElement();

    byteData.at(byteData.size() - 1)[endFilledBytes] = source;
    endFilledBytes++;
}//appendByte

void DBytes::insertByte(uint8_t source)
{
    if (startFilledBytes >= ARRAY_SIZE)
    {
        if (byteData.size() == 1)
            endFilledBytes = ARRAY_SIZE;

        #ifdef DA_DEBUG_MODE
            byteData.push_front(DSafeArray(ARRAY_SIZE, true));
            byteData.at(0).makePermanent();
            arraysAdded++;
        #else
            byteData.push_front(new uint8_t[ARRAY_SIZE]);
        #endif

        startFilledBytes = 0;
    }//if startFilledBytes

    byteData.at(0)[ARRAY_SIZE - (startFilledBytes + 1)] = source;

    startFilledBytes++;
}//insertByte

void DBytes::removeFirst(long int numRemovals)
{
    uint32_t completeElements = 0;//the number of full elements in byteData to delete
    const long int origSize = size();//the original number of bytes in byteData
    const long int origNumRemovals = numRemovals;

    if (numRemovals >= static_cast<long int>(size()))
        clear();
    else
    {
        if (numRemovals < startFilledBytes)
            startFilledBytes -= numRemovals;
        else
        {
            //### Remove the first element ###
            numRemovals -= startFilledBytes;
            if (origNumRemovals >= startFilledBytes)
                startFilledBytes = 0;
            else
                startFilledBytes -= origNumRemovals;

            //### Remove complete elements ###
            if (numRemovals > 0)
                completeElements = numRemovals / ARRAY_SIZE;

            for (unsigned int i = 0; i < completeElements; i++)
            {
                numRemovals -= ARRAY_SIZE;

                #ifndef DA_DEBUG_MODE
                    delete byteData.at(0);
                #else
                    arraysAdded--;
                #endif
                byteData.pop_front();
            }//for i

            //### Prune the final element ###
            if (numRemovals > 0)
            {
                //Should be less than one element's worth of bytes to remove
                if (numRemovals >= ARRAY_SIZE)
                    DEngine::fatalError("DBytes::removeFirst(): numRemovals > ARRAY_SIZE");

                //We already pruned the first element, something's wrong
                if (byteData.size() <= 1)
                    DEngine::fatalError("DBytes::removeFirst(): byteData.size() <= 1");

                //Remove the first element, which should have zero bytes
                else
                {
                    if (startFilledBytes != 0)
                        DEngine::fatalError("DBytes::removeFirst(): startFilledBytes == 0");

                    #ifndef DA_DEBUG_MODE
                        delete byteData.at(0);
                    #else
                        arraysAdded--;
                    #endif
                    byteData.pop_front();
                }//else size()

                //Configure a new first element
                if (byteData.size() == 1)
                {
                    if (numRemovals > endFilledBytes)
                        DEngine::fatalError("DBytes::removeFirst(): numRemovals > endFilledBytes");

                    convertLastToFirst(numRemovals);
                }//if size()
                else
                    startFilledBytes = ARRAY_SIZE - numRemovals;
            }//if numRemovals
        }//else numRemovals
    }//else numRemovals

    if (static_cast<long int>(size()) != origSize - origNumRemovals)
    {
        DEngine::addFatal("DBytes::removeFirst(): removed wrong number of bytes! size: ");
        DEngine::addFatal(DEngine::intToStr(size()));
        DEngine::addFatal(" expected size: ");
        DEngine::fatalError(DEngine::intToStr(origSize - origNumRemovals));
    }//if size()
}//removeFirst

void DBytes::convertLastToFirst(long int numRemovals)
{
    #ifdef DA_DEBUG_MODE
        DSafeArray oneElement(0, false);//the only element in dataBytes
    #else
        uint8_t *oneElement = NULL;//the only element in dataBytes
    #endif
    unsigned int sourceStart = 0;//the indexes of the first and last array items to copy
    unsigned int sourceEnd = 0;
    unsigned int writeIndex = 0;//the index in oneElement to write to

    //### Calculate the number of bytes in the first element ###
    startFilledBytes = endFilledBytes - numRemovals;
    if (startFilledBytes > ARRAY_SIZE - numRemovals)
        DEngine::fatalError("DBytes::convertLastToFirst(): startFilledBytes too large");

    //### Zero endFilledBytes ###
    endFilledBytes = 0;

    //### Extract the first element ###
    oneElement = byteData.at(0);

    //### Calculate the range of bytes to copy ###
    sourceStart = startFilledBytes + numRemovals - 1;
    sourceEnd = numRemovals;

    //### Start writing at the bottom of the array ###
    writeIndex = ARRAY_SIZE - 1;

    //### Shuffle the bytes ###
    for (unsigned int i = sourceStart; i >= sourceEnd; i--)
    {
        oneElement[writeIndex] = oneElement[i];
        writeIndex--;
    }//for i
}//convertLastToFirst

void DBytes::append(const DBytes &source)
{
    append(source, 0, source.size());
}//append

void DBytes::append(const DBytes &source,
                    long int index,
                    long int length)
{
    #ifdef DA_DEBUG_MODE
        DSafeArray readBuffer(0, false);//the element from the source to read from
        DSafeArray writeBuffer(0, false);//the element in byteData to write to
    #else
        uint8_t *readBuffer = NULL;//the element from the source to read from
        uint8_t *writeBuffer = NULL;//the element in byteData to write to
    #endif
    uint32_t sourceElement = 0;//the first element in source to copy
    unsigned int sourceArray = 0;//the first array index to copy
    unsigned int writeAmount = 0;//the number of bytes to copy
    unsigned int destCapacity = 0;//the number of free bytes in writeElement

    //### Find the index of the first element to copy ###
    source.calcIndexes(index, sourceElement, sourceArray);

    //### Create a new element ###
    if (byteData.size() == 0)
    {
        startFilledBytes = 0;
        addElement();
    }//if size()

    if (byteData.size() == 1)
    {
        addElement();
    }//if size()

    while ((sourceElement < source.byteData.size()) && (length > 0))
    {
        //### Verify that the parameters are valid ###
        if (index < 0)
        {
            DEngine::addFatal("DBytes::append(): index is ");
            DEngine::fatalError(DEngine::intToStr(index));
        }//if index
        if (static_cast<unsigned int>(index + length) > source.size())
        {
            DEngine::addFatal("DBytes::append(): index + length is ");
            DEngine::addFatal(DEngine::intToStr(index + length));
            DEngine::addFatal(", size is ");
            DEngine::fatalError(DEngine::intToStr(size()));
        }//if index + length

        //### Find the index of the first element to copy ###
        source.calcIndexes(index, sourceElement, sourceArray);

        //### Calculate the number of bytes we can copy from the source ###
        if (sourceElement == source.byteData.size() - 1)
            writeAmount = source.endFilledBytes - sourceArray;
        else
            writeAmount = ARRAY_SIZE - sourceArray;

        //### Calculate the number of bytes we can write to the dest ###
        destCapacity = ARRAY_SIZE - endFilledBytes;
        if (destCapacity == 0)
        {
            addElement();
            destCapacity = ARRAY_SIZE;
        }//if destCapacity

        //### Determine the final number of bytes to copy ###
        if (writeAmount > destCapacity)
            writeAmount = destCapacity;
        if (writeAmount > static_cast<unsigned long int>(length))
            writeAmount = length;

        //### Extract the arrays to read and write to ###
        readBuffer = source.byteData.at(sourceElement);
        writeBuffer = byteData.at(byteData.size() - 1);

        //### Perform the copy ###
        #ifdef DA_DEBUG_MODE
            for (unsigned int i = endFilledBytes; i < endFilledBytes + writeAmount; i++)
            {
                writeBuffer[i] = readBuffer[sourceArray];
                sourceArray++;
            }//for i
        #else
            memcpy(&writeBuffer[endFilledBytes], &readBuffer[sourceArray], writeAmount);
        #endif

        endFilledBytes += writeAmount;
        index += writeAmount;
        length -= writeAmount;
    }//while sourceElement && length

    if (length != 0)
    {
        DEngine::addFatal("DBytes::append(): bytes remaining to append: ");
        DEngine::fatalError(DEngine::intToStr(length));
    }//if length
}//append

void DBytes::appendArray(const uint8_t *source,
                         long int length)
{
    #ifdef DA_DEBUG_MODE
        DSafeArray writeBuffer(0, false);//the element in byteData to write to
    #else
        uint8_t *writeBuffer = NULL;//the element in byteData to write to
    #endif
    long int sourceIndex = 0;//the index of the first byte in source to copy
    unsigned int writeAmount = 0;//the number of bytes we can write to the destination buffer
    unsigned int destCapacity = 0;//the number of free bytes in writeElement

    //### Create a new element if needed ###
    if (byteData.size() == 0)
    {
        startFilledBytes = 0;
        addElement();
    }//if size()

    if (byteData.size() == 1)
        addElement();

    while (sourceIndex < length)
    {
        //### Create a new element if necessary ###
        destCapacity = ARRAY_SIZE - endFilledBytes;
        if (destCapacity == 0)
        {
            addElement();
            destCapacity = ARRAY_SIZE;
        }//if destCapacity

        //### Calculate the number of bytes we can copy to the dest ###
        writeAmount = length - sourceIndex;
        if (writeAmount > destCapacity)
            writeAmount = destCapacity;

        //### Copy the data ###
        writeBuffer = byteData.at(byteData.size() - 1);

        //cout << "endFilledBytes: " << endFilledBytes << " writeAmount: " << writeAmount << " ARRAY_SIZE: " << ARRAY_SIZE << endl;

        #ifdef DA_DEBUG_MODE
            for (unsigned int i = endFilledBytes; i < endFilledBytes + writeAmount; i++)
            {
                writeBuffer[i] = source[sourceIndex];
                sourceIndex++;
            }//for i
        #else
            memcpy(&writeBuffer[endFilledBytes], &source[sourceIndex], writeAmount);
            sourceIndex += writeAmount;
        #endif

        endFilledBytes += writeAmount;
    }//while sourceIndex

    if (sourceIndex != length)
    {
        DEngine::addFatal("DBytes::appendArray(): copied ");
        DEngine::addFatal(DEngine::intToStr(sourceIndex));
        DEngine::addFatal(" out of ");
        DEngine::fatalError(DEngine::intToStr(length));
    }//if sourceIndex
}//appendArray

vector<boost::asio::const_buffer> DBytes::getSendBuffers(long int maxBytes) const
{
    vector<boost::asio::const_buffer> result;//the buffers to return
    unsigned int numFromCurrent = 0;//the number of bytes to copy from the current element
    #ifdef DA_DEBUG_MODE
        DSafeArray dataElement(0, false);//the current buffer to send
    #else
        uint8_t *dataElement = NULL;//the current buffer to send
    #endif

    if (byteData.size() > 0)
    {
        //### Append the first buffer ###
        if (startFilledBytes > 0)
        {
            numFromCurrent = startFilledBytes;
            if (maxBytes < startFilledBytes)
                numFromCurrent = maxBytes;

            dataElement = byteData.at(0);
            result.push_back(boost::asio::buffer(&dataElement[ARRAY_SIZE - startFilledBytes],
                                                 numFromCurrent));

            maxBytes -= numFromCurrent;
        }//if startFilledBytes

        //### Append the filled buffers ###
        for (unsigned long int i = 1; (i < byteData.size() - 1) && (maxBytes > 0); i++)
        {
            dataElement = byteData.at(i);

            numFromCurrent = ARRAY_SIZE;
            if (maxBytes < ARRAY_SIZE)
                numFromCurrent = maxBytes;

            result.push_back(boost::asio::buffer(&dataElement[0], numFromCurrent));
            maxBytes -= numFromCurrent;
        }//for i

        //### Append the last buffer ###
        if (byteData.size() > 1)
        {
            if (endFilledBytes > 0)
            {
                numFromCurrent = endFilledBytes;
                if (maxBytes < endFilledBytes)
                    numFromCurrent = maxBytes;

                dataElement = byteData.at(byteData.size() - 1);

                result.push_back(boost::asio::buffer(&dataElement[0], numFromCurrent));

                maxBytes -= numFromCurrent;
            }//if endFilledBytes
        }//if size()
    }//if size()

    return result;
}//getSendBuffers

void DBytes::addElement(void)
{
    #ifdef DA_DEBUG_MODE
        byteData.push_back(DSafeArray(ARRAY_SIZE, true));
        byteData.at(byteData.size() - 1).makePermanent();
        arraysAdded++;
    #else
        byteData.push_back(new uint8_t[ARRAY_SIZE]);
    #endif

    endFilledBytes = 0;
}//addElement

long int DBytes::getNumInstances(void)
{
    #ifdef DA_DEBUG_MODE
        return numInstances;
    #else
        return 0;
    #endif
}//getNumInstances
