/*
 * Copyright (C) 2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DClient.hpp"
#include "DEngine.hpp"
#include "DBuffer.hpp"

using namespace std;
using namespace libdasyne;

DClient::DClient(DEngine *newNetEngine)
{
    netEngine = newNetEngine;
    clientSocket = new boost::asio::ip::tcp::socket(*netEngine->getIoService());
    bufferSize = DEFAULT_BUFFER_SIZE;
    
    connected = false;
    initVars();
}//constructor

DClient::DClient(DEngine *newNetEngine,
                 unsigned int newBufferSize)
{
    netEngine = newNetEngine;
    clientSocket = new boost::asio::ip::tcp::socket(*netEngine->getIoService());

    bufferSize = newBufferSize;
    if (bufferSize < 1024)
        bufferSize = 1024;

    connected = false;
    initVars();
}//constructor

DClient::DClient(DEngine *newNetEngine,
                 boost::asio::ip::tcp::socket *newSocket)
{
    netEngine = newNetEngine;
    clientSocket = newSocket;
    bufferSize = DEFAULT_BUFFER_SIZE;

    checkConnected();
    initVars();
    startReceive();
}//constructor

DClient::DClient(DEngine *newNetEngine,
                 boost::asio::ip::tcp::socket *newSocket,
                 unsigned int newBufferSize)
{
    netEngine = newNetEngine;
    clientSocket = newSocket;

    bufferSize = newBufferSize;
    if (bufferSize < 1024)
        bufferSize = 1024;

    checkConnected();
    initVars();
    startReceive();
}//constructor

DClient::~DClient(void)
{
    delete sendQueue;
    sendQueue = NULL;

    delete clientSocket;
    clientSocket = NULL;

    delete hostResolver;
    hostResolver = NULL;

    delete [] receiveBuffer;
    receiveBuffer = NULL;
}//destructor

void DClient::checkConnected(void)
{
    boost::system::error_code errorCode;//error code returned by boost
    
    connected = false;
    if (clientSocket->is_open())
    {
        clientSocket->remote_endpoint(errorCode);
        if (!errorCode)
            connected = true;
    }//if is_open()
}//checkConnected

void DClient::initVars(void)
{
    hostResolver = new boost::asio::ip::tcp::resolver(*netEngine->getIoService());
    receiveBuffer = new unsigned char[bufferSize];
    sendQueue = new DBuffer;
    endPointIter = boost::asio::ip::tcp::resolver::iterator();
    isSending = false;
    delayEnabled = true;
}//initVars

void DClient::startConnect(const string &hostName,
                           uint16_t newPortNumber)
{
    boost::asio::ip::tcp::resolver::query *nameQuery = NULL;//query to resolve hostName to an IP address
    string portString = DEngine::intToStr(newPortNumber);

    clientSocket->close();
    connected = false;
    
    nameQuery = new boost::asio::ip::tcp::resolver::query(hostName, portString);

    hostResolver->async_resolve(*nameQuery, boost::bind(&DClient::resolveHandler, this,
                                                        boost::asio::placeholders::error,
                                                        boost::asio::placeholders::iterator));

    delete nameQuery;
    nameQuery = NULL;
}//startConnect

void DClient::resolveHandler(const boost::system::error_code &errorCode,
                             boost::asio::ip::tcp::resolver::iterator iterator)
{
    boost::asio::ip::tcp::endpoint endpoint;//the IP address and port number to connect to

    endPointIter = iterator;

    if (errorCode)
    {
        errMsg = "Couldn't resolve hostname: ";
        errMsg.append(errorCode.message());
        connectHandler(errMsg, DA_HOSTNAME_ERROR);
    }//if error
    else
    {
        endpoint = *endPointIter;
        endPointIter++;

        clientSocket->async_connect(endpoint, boost::bind(&DClient::connectFinished, this,
                                                          boost::asio::placeholders::error));
    }//else error
}//resolveHandler

void DClient::connectFinished(const boost::system::error_code &errorCode)
{
    boost::asio::ip::tcp::endpoint endpoint;//the IP address and port number to connect to
    bool allOK = true;//false on networking error

    //### Try a different endpoint if we couldn't connect ###
    if (errorCode)
    {
        if (errorCode != boost::asio::error::operation_aborted)
        {
            //Verify that there's another endpoint
            if (endPointIter == boost::asio::ip::tcp::resolver::iterator())
            {
                allOK = false;
                errMsg = "Couldn't connect: ";
                errMsg.append(errorCode.message());
                connectHandler(errMsg, DA_CONNECT_ERROR);
            }//if endPointer

            //Close the socket
            if (allOK)
            {
                if (!closeSocket())
                {
                    allOK = false;
                    connectHandler(errMsg, DA_CLOSE_ERROR);
                }//if !closeSocket()
            }//if allOK

            //Start another asynchronous connect
            if (allOK)
            {
                endpoint = *endPointIter;
                endPointIter++;

                clientSocket->async_connect(endpoint, boost::bind(&DClient::connectFinished, this,
                                                                  boost::asio::placeholders::error));
            }//if allOK
        }//if errorCode
    }//if errorCode

    //### We connected successfully ###
    else
    {
        connected = true;
        startReceive();
        connectHandler("Success", DA_SUCCESS);
    }//else errorCode
}//connectFinished

void DClient::send(const DBuffer &theData)
{
    sendQueue->append(theData);

    sendBytes();
}//send

void DClient::sendBytes(void)
{
    vector<boost::asio::const_buffer> sendBuffers;//the buffers to send
    int maxBytes = bufferSize;//the maximum number of bytes to send

    if (!isSending)
    {
        //### Find the number of bytes to send ###
        if (sendQueue->size() < bufferSize)
            maxBytes = sendQueue->size();

        //### Send the data ###
        if (maxBytes > 0)
        {
            sendBuffers = sendQueue->getSendBuffers(maxBytes);

            isSending = true;
            clientSocket->async_send(sendBuffers,
                                     boost::bind(&DClient::sendFinished, this,
                                     boost::asio::placeholders::error,
                                     boost::asio::placeholders::bytes_transferred));
        }//if maxBytes
        else
            sendHandler();
    }//if !isSending
}//sendBytes

unsigned long int DClient::unsentBytes(void)
{
    return sendQueue->size();
}//unsentBytes

void DClient::sendFinished(const boost::system::error_code &errorCode,
                           size_t bytesSent)
{
    //### Verify that we're sending something ###
    if (!isSending)
        netEngine->fatalError("DClient::sendFinished(): called, but we're not sending!");

    if (bytesSent >= 1)
        sendQueue->removeFirst(bytesSent);
    else
    {
        if (clientSocket->is_open())
        {
            connected = false;
            errMsg = "Sent zero bytes";
            errorHandler(errMsg);
            closeSocket();
        }//if is_open()
    }//else bytesSent

    isSending = false;

    if (errorCode)
    {
        if (errorCode != boost::asio::error::operation_aborted)
            reportError(errorCode);
    }//if errorCode
    else
        sendBytes();
}//sendFinished

bool DClient::closeSocket(void)
{
    boost::system::error_code errorCode;//error code returned by an ASIO operation
    boost::system::error_condition theError;//convert errorCode into a portable value
    bool result = true;//false if we couldn't close the socket

    connected = false;

    if (clientSocket->is_open())
    {
        //### Shutdown the socket ###
        clientSocket->shutdown(boost::asio::ip::tcp::socket::basic_stream_socket::shutdown_both, errorCode);
        if (errorCode)
        {
            theError = errorCode.default_error_condition();

            if (theError.value() != boost::system::errc::not_connected)
            {
                errMsg = "Couldn't shutdown socket: ";
                errMsg.append(theError.message());
                result = false;
            }//if value()
        }//if errorCode

        //### Close the socket ###
        if (result)
        {
            clientSocket->close(errorCode);
            if (errorCode)
            {
                theError = errorCode.default_error_condition();

                errMsg = "Couldn't close socket: ";
                errMsg.append(theError.message());
                result = false;
            }//if errorCode
        }//if result
    }//if is_open()

    return result;
}//closeSocket

void DClient::connectHandler(std::string message,
                             NetworkCode status)
{
}//connectHandler

void DClient::sendHandler(void)
{
}//sendHandler

void DClient::startReceive(void)
{
    clientSocket->async_receive(boost::asio::buffer(receiveBuffer, bufferSize),
                                boost::bind(&DClient::receiveFinished, this,
                                            boost::asio::placeholders::error,
                                            boost::asio::placeholders::bytes_transferred));
}//startReceive

void DClient::receiveFinished(const boost::system::error_code &errorCode,
                              size_t bytesReceived)
{
    DBuffer receivedData;//the data received

    if (errorCode)
    {
        if (errorCode != boost::asio::error::operation_aborted)
            reportError(errorCode);
    }//if errorCode
    else
    {
        receivedData.appendArray(receiveBuffer, bytesReceived);

        if (receivedData.size() > 0)
            receiveHandler(receivedData);
        else
        {
            if (clientSocket->is_open())
            {
                connected = false;
                errMsg = "Received zero bytes.";
                errorHandler(errMsg);
                closeSocket();
            }//if is_open()
        }//else size()

        startReceive();
    }//else
}//receiveFinished

string DClient::getLocalAddress(void)
{
    boost::system::error_code errorCode;//an error code from calling local_endpoint()
    boost::system::error_condition theError;//convert errorCode to a portable value
    boost::asio::ip::tcp::endpoint localEndpoint;//the local endpoint for this DClient
    boost::asio::ip::address localAddress;//the local address of this DClient
    string result;//the address to return

    localEndpoint = clientSocket->local_endpoint(errorCode);
    if (errorCode)
    {
        theError = errorCode.default_error_condition();
        errMsg = theError.message();
        errMsg.append(": ");
        errMsg.append(DEngine::intToStr(theError.value()));
    }//if errorCode
    else
    {
        localAddress = localEndpoint.address();
        result = localAddress.to_string();
    }//else errorCode

    return result;
}//getLocalAddress

std::string DClient::getRemoteAddress(void)
{
    boost::system::error_code errorCode;//an error code from calling local_endpoint()
    boost::system::error_condition theError;//convert errorCode to a portable value
    boost::asio::ip::tcp::endpoint remoteEndpoint;//the remote host's endpoint
    boost::asio::ip::address remoteAddress;//the remote host's IP address
    string result;//the address to return

    if (clientSocket->is_open())
    {
        remoteEndpoint = clientSocket->remote_endpoint(errorCode);
        if (errorCode)
        {
            theError = errorCode.default_error_condition();
            errMsg = theError.message();
            errMsg.append(": ");
            errMsg.append(DEngine::intToStr(theError.value()));
        }//if errorCode
        else
        {
            remoteAddress = remoteEndpoint.address();
            result = remoteAddress.to_string();
        }//else errorCode
    }//if is_open()

    return result;
}//getRemoteAddress

void DClient::receiveHandler(DBuffer &receivedBytes)
{
}//receiveHandler

void DClient::reportError(const boost::system::error_code &errorCode)
{
    boost::system::error_condition theError = errorCode.default_error_condition();

    connected = false;

    if (clientSocket->is_open())
    {
        errMsg = theError.message();
        errMsg.append(": ");
        errMsg.append(DEngine::intToStr(theError.value()));

        errorHandler(errMsg);
        closeSocket();
    }//if is_open()
}//reportError

bool DClient::setDelay(bool newValue)
{
    boost::asio::ip::tcp::no_delay option(!newValue);
    boost::system::error_code errorCode;//ASIO error code returned by set_option
    bool result = true;//false if we couldn't set the option

    clientSocket->set_option(option, errorCode);
    if (errorCode)
    {
        errMsg = "Couldn't set option: ";
        errMsg.append(errorCode.message());
        result = false;
    }//if errorCode
    else
        delayEnabled = newValue;

    delayToggled();

    return result;
}//setDelay

bool DClient::getDelayEnabled(void)
{
    return delayEnabled;
}//getDelayEnabled

bool DClient::isConnected(void)
{
    return connected;
}//isConnected

void DClient::errorHandler(std::string message)
{

}//errorHandler

void DClient::delayToggled(void)
{
}//delayToggled

unsigned int DClient::getBufferSize(void)
{
    return bufferSize;
}//getBufferSize

string DClient::getErrMsg(void)
{
    return errMsg;
}//getErrMsg
