/*
 * Copyright (C) 2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    Handles events for other Dasyne classes. DEngine contains an update function
    that should be called periodically, and dispatches events to objects belonging
    to it. Other classes accept the DEngine in their constructors, and will receive
    events from it. All events are single-threaded, they're called one at a time when
    the update function is run.
*/

#ifndef DENGINE_HPP_INCLUDED
#define DENGINE_HPP_INCLUDED

#ifdef _MSC_VER
    #include <inttypes.h>
#endif

#ifdef DA_WINDOWS_VERSION
    #ifndef _WIN32_WINDOWS
        #ifndef _WIN32_WINNT
            //#warning _WIN32_WINDOWS not defined, defaulting to 0x0501
            //#warning Codes: 0x0501 = Windows XP, 0x0600 = Windows Vista, 0x0601 = Windows 7
            #define _WIN32_WINDOWS 0x0501
        #endif
    #endif

    #include "zlibheaders/zlib.h"
#else
    #include "zlib.h"
#endif

#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/function.hpp>

#include <iostream>
#include <string>
#include <sstream>
#include <list>

namespace libdasyne
{
    enum EndianType
    {
        DA_UNKNOWN, DA_BIG_ENDIAN, DA_LITTLE_ENDIAN
    };//EndianType

    enum SocketAction
    {
        DASA_NONE, DASA_SHOW_MESSAGE, DASA_CLOSE_SOCKET, DASA_UNKNOWN
    };//SocketAction

    enum IpVersion
    {
        DA_IPV4 = 4,
        DA_IPV6 = 6
    };//IPVersion

    enum NetworkCode
    {
        DA_SUCCESS = 0,
        DA_HOSTNAME_ERROR = -1, //couldn't resolve the host name
        DA_CLOSE_ERROR =    -2, //the socket was already open and we couldn't close it
        DA_CONNECT_ERROR =  -3, //couldn't connect to the specified host
    };//NetworkCode

    class DTime;
    class DEvent;
    class DMutex;

    class DEngine
    {
        public:
            //constructor
            DEngine(void);

            //destructor
            ~DEngine(void);

            //Dispatches queued events for objects using this network engine. Returns -1
            //on error and sets errMsg, otherwise returns the number of events we dispatched.
            //maxEvents is the maximum number of events to dispatch, or -1 for all.
            int update(int maxEvents);

            //Returns a pointer to the boost io_service, which links other classes to this
            //DEngine's event handlers.
            boost::asio::io_service* getIoService(void);

            //Report that an unrecoverable error has occurred, and throw an exception
            static void addFatal(const std::string &description);
            static void fatalError(const std::string &description);

            //returns a description of the most recent error
            std::string getErrMsg(void);

            //Adds newEvent to our event list, it's update function will be called whenever
            //DEngine::update is called. The DEngine does not take ownership of it.
            std::list<DEvent*>::iterator addEvent(DEvent *newEvent);

            //Removes the event at the specified location in eventList, does not delete it.
            void removeEvent(std::list<DEvent*>::iterator eventIter);

            //Adds newTime to our time list. It will be notified if this DEngine is destroyed.
            std::list<DTime*>::iterator addTime(DTime *newTime);

            //Removes the DTime at the specified location in timeList, does not delete it.
            void removeTime(std::list<DTime*>::iterator timeIter);

            //Sets seconds and milliseconds to the current time. The time is updated whenever
            //update() is called
            void getCurrentTime(uint32_t &seconds,
                                uint16_t &milliseconds);

            //Suspends execution of the calling thread by the specified number of milliseconds
            static void sleep(long int milliseconds);

            //returns the string equivalent of number
            static std::string intToStr(long int number);

            //returns false if text is not a valid integer
            static bool strToInt(const std::string &text,
                                 int &theInt);

            //Returns the endianness for this machine
            static libdasyne::EndianType getEndianness(void);

            //Outputs the error code for each error in errc. There doesn't seem to be
            //any other way to match an error_code value to its meaning.
            static void outputErrorCodes(void);

        private:
            //A list of non-ASIO event-classes, we'll call their update functions whenever
            //DEngine::update is called.
            std::list<DEvent*> eventList;

            //A list of DTime objects that will be notified when this DEngine is destroyed.
            std::list<DTime*> timeList;

            //The current event in eventList to update
            std::list<DEvent*>::iterator updateIter;

            //Our link to the operating system's IO services
            boost::asio::io_service *ioService;

            //Other DTime objects can get their time from globalTime, reducing the number
            //of expensive system calls. globalTime is updated whenever update() is called
            DTime *globalTime;

            //Controls access to fatalMsg
            static libdasyne::DMutex *fatalMutex;

            //a description of the most recent error
            std::string errMsg;

            //An error message related to a fatal error that caused an exception to be thrown
            static std::string fatalMsg;

            //The endianness of this machine
            static libdasyne::EndianType endianness;

            //The number of DEngine objects
            static int numEngines;

            //True if the current DEvent should be erased once it's update method finishes. Prevents
            //crashes when a DEvent removes itself from the eventList during it's update function.
            bool havePendingDeletion;

            //############### Functions ###############

            //Calculate the type of endianness used on this machine
            static libdasyne::EndianType calcEndianness(void);

    };//DEngine
}//namespace libdasyne

#endif
