/*
 * Copyright (C) 2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DServer.hpp"
#include "DEngine.hpp"
#include "DClient.hpp"

using namespace std;
using namespace libdasyne;

DServer::DServer(DEngine *newNetEngine)
{
    netEngine = newNetEngine;

    listenSocket = new boost::asio::ip::tcp::acceptor(*netEngine->getIoService());
    clientSocket = NULL;
    myState = DA_UNBOUND;
}//constructor

DServer::~DServer(void)
{
    delete listenSocket;
    listenSocket = NULL;

    delete clientSocket;
    clientSocket = NULL;
}//destructor

bool DServer::bind(libdasyne::IpVersion ipVersion,
                   uint16_t portNumber)
{
    boost::asio::ip::tcp::acceptor::reuse_address reuseAddr(true);//enable address re-use
    boost::system::error_code errorCode;//error code returned by an ASIO operation
    boost::asio::ip::tcp::acceptor::endpoint_type *endPoint = NULL;//the protocol and port number to bind to
    bool result = true;//false on network errors

    //### Close the listen socket if open ###
    if (listenSocket->is_open())
        listenSocket->close();
    myState = DA_UNBOUND;

    //### Open the socket ###
    if (ipVersion == libdasyne::DA_IPV4)
        listenSocket->open(boost::asio::ip::tcp::v4());
    else if (ipVersion == libdasyne::DA_IPV6)
        listenSocket->open(boost::asio::ip::tcp::v6());
    else
    {
        result = false;
        errMsg = "Unknown IP version ";
        errMsg.append(DEngine::intToStr(ipVersion));
    }//else ipVersion

    //### Allow address reuse ###
    if (result)
    {
        listenSocket->set_option(reuseAddr, errorCode);
        if (errorCode)
        {
            result = false;
            errMsg = "Couldn't enable address reuse: ";
            errMsg.append(errorCode.message());
        }//if errorCode
    }//if result

    //### Bind the socket ###
    if (result)
    {
        if (ipVersion == libdasyne::DA_IPV4)
            endPoint = new boost::asio::ip::tcp::acceptor::endpoint_type(boost::asio::ip::tcp::v4(), portNumber);
        else
            endPoint = new boost::asio::ip::tcp::acceptor::endpoint_type(boost::asio::ip::tcp::v6(), portNumber);

        listenSocket->bind(*endPoint, errorCode);

        delete endPoint;
        endPoint = NULL;

        if (errorCode)
        {
            result = false;
            errMsg = "Couldn't bind the socket: ";
            errMsg.append(errorCode.message());
        }//if errorCode
        else
            myState = DA_BOUND;
    }//if result

    return result;
}//bind

bool DServer::startListening(void)
{
    boost::system::error_code errorCode;//error code returned by an ASIO operation
    bool result = true;//false if we couldn't start listening

    //### Verify that the socket is in the BOUND state ###
    if (myState == DA_UNBOUND)
    {
        errMsg = "The socket must be bound before startListening is called.";
        result = false;
    }//if myState
    else if (myState == DA_LISTENING)
    {
        errMsg = "The socket is already listening.";
        result = false;
    }//else if myState

    //### Enable listening ###
    if (result)
    {
        listenSocket->listen(boost::asio::socket_base::max_connections, errorCode);
        if (errorCode)
        {
            result = false;
            errMsg = "Couldn't listen for connections: ";
            errMsg.append(errorCode.message());
        }//if errorCode
        else
        {
            myState = DA_LISTENING;
            getNextConnection();
        }//else errorCode
    }//if result

    return result;
}//startListening

void DServer::connectFinished(const boost::system::error_code &errorCode)
{
    boost::system::error_condition theError;//convert errorCode into a portable value

    if (errorCode)
    {
        theError = errorCode.default_error_condition();

        if ((theError != boost::asio::error::operation_aborted)
         && (theError.value() != boost::system::errc::bad_file_descriptor))
        {
            errMsg = theError.message();
            errMsg.append(": ");
            errMsg.append(DEngine::intToStr(theError.value()));

            errorHandler(errMsg);
            listenSocket->close();
        }//if theError && value()

        delete clientSocket;
        clientSocket = NULL;
    }//if errorCode

    else
    {
        connectHandler(clientSocket);
        clientSocket = NULL;

        getNextConnection();
    }//else errorCode
}//connectFinished

void DServer::getNextConnection(void)
{
    if (clientSocket != NULL)
        netEngine->fatalError("DServer::startListening(): clientSocket is not NULL!");
    clientSocket = new boost::asio::ip::tcp::socket(*netEngine->getIoService());

    listenSocket->async_accept(*clientSocket, clientAddress,
                               boost::bind(&DServer::connectFinished, this, boost::asio::placeholders::error));
}//getNextConnection

void DServer::close(void)
{
    listenSocket->close();
}//close

string DServer::getErrMsg(void)
{
    return errMsg;
}//getErrMsg
